from importlib.abc import Traversable
from importlib.resources import files
import logging
from typing import Any


class Config:
    def __init__(self, config_dir: Traversable, filename: str):
        self._path = config_dir.joinpath(filename)
        self._data = dict()

    def get(self, key, default=None) -> Any:
        return self._data.get(key, default)

    def set(self, key, value):
        self._data[key] = value

    def __getitem__(self, key) -> Any:
        return self.get(key)

    def __setitem__(self, key, value):
        self.set(key, value)


logging.basicConfig(level=logging.INFO)
log = logging.getLogger("live_template")
log.setLevel(logging.INFO)
package_root = files("live_template")
config_dir = package_root.joinpath("config")
config = Config(config_dir, "config.json")



from abc import ABC, abstractmethod
import asyncio
from asyncio import AbstractEventLoop
from pathlib import Path

from ..core import config, package_root
from ..storage.storage import (
    InlineButton,
    InternalTemplateStorage,
    Template,
    TemplateStorage,
)
from ..watcher.watcher import TemplateWatcher


class BaseRouter(ABC):
    TEMPLATE_CALLBACK_ID = "template"
    START_COMMAND = "start_lt"
    TEMPLATE_CALLBACK_SEP = ":"

    def __init__(self, templates_dir: str | Path, default_template: dict):
        super().__init__()

        self.command_handlers = {}
        self.callback_handlers = {}

        self._templates_dir = Path(templates_dir)
        self._default_template = default_template

        self._internal_templates_dir = package_root / "core" / "templates"
        self._internal_default_template = {"text": "Default template"}
        self._internal_storage = InternalTemplateStorage(
            self._internal_templates_dir, self._internal_default_template
        )

        self._storage = TemplateStorage(self._templates_dir, self._default_template)

        self._tw_queue = asyncio.Queue()
        self._tw_loop: AbstractEventLoop | None = None
        self._tw: TemplateWatcher | None = None

        self._chat_id = None
        self._always_retry = False

        self.setup_handlers()

    def setup_handlers(self):
        # self.register_command("lt_start")(self._start)
        self.register_command("list_template_names")(self._list_template_names)

        self.register_callback("list_template_names")(self._list_template_names)
        self.register_callback("toggle_always_retry")(self._toggle_always_retry)
        self.register_callback("list_templates")(self._list_templates)

        # self.register_callback(self._template_callback_id)(self._get_template)

    def register_command(self, command: str):
        def wrapper(func):
            self.command_handlers[command] = func
            return func

        return wrapper

    def register_callback(self, callback: str):
        def wrapper(func):
            self.callback_handlers[callback] = func
            return func

        return wrapper

    # async def handle(self, command: str, *args, **kwargs):
    #     handler = self.handlers.get(command)
    #     if handler:
    #         if inspect.iscoroutinefunction(handler):
    #             return await handler(*args, **kwargs)
    #         else:
    #             loop = asyncio.get_event_loop()
    #             return await loop.run_in_executor(None, lambda: handler(*args, **kwargs))

    @abstractmethod
    async def _send_msg(self, chat_id: int, template: Template):
        pass

    def _template(
        self, template_name: str, is_internal: bool = False
    ) -> Template | None:
        storage = self._internal_storage if is_internal else self._storage
        template = storage[template_name]
        if not template:
            return None

        return template

    @staticmethod
    def _template_callback(template_name: str) -> str:
        return (
            BaseRouter.TEMPLATE_CALLBACK_ID
            + BaseRouter.TEMPLATE_CALLBACK_SEP
            + template_name
        )

    def _on_startup(self):
        self._tw_loop = asyncio.get_running_loop()
        self._tw = TemplateWatcher(self._storage, self._tw_queue, self._tw_loop)
        self._tw.start_watching()
        asyncio.create_task(self._always_retry_dispatcher())

    async def _on_shutdown(self):
        await self._tw_queue.put(None)

    # TODO: узкое место: chat_id получается только после нажатия /start
    # TODO: небезопасно: пользователи могут получить доступ к шаблонам во время разработки
    # Command: lt_start
    async def _start(self, chat_id: int):
        self._chat_id = chat_id
        template = self._template("live_template_start", is_internal=True)
        await self._send_msg(self._chat_id, template)

    # Command: list_template_names
    # Callback: list_template_names
    async def _list_template_names(self, *args, **kwargs):
        template_names = sorted(self._storage.names())
        template_name = "list_templates" if template_names else "no_templates"
        template = self._template(template_name, is_internal=True)
        template.buttons = []
        for name in template_names:
            template.buttons.append(InlineButton(name, self._template_callback(name)))

        await self._send_msg(self._chat_id, template)

    # Callback: toggle_always_retry
    async def _toggle_always_retry(self, *args, **kwargs):
        self._always_retry = not self._always_retry
        template = self._template(
            f"always_retry_{str(self._always_retry).lower()}", is_internal=True
        )
        config["always_retry"] = self._always_retry
        await self._send_msg(self._chat_id, template)

    # Callback: list_templates
    async def _list_templates(self, *args, **kwargs):
        for name, template in sorted(self._storage.list()):
            await self._send_msg(self._chat_id, Template(name=name, text=name))
            await self._send_msg(self._chat_id, template)

    # Callback: self._template_callback(template_name)
    async def _get_template(self, template_name: str):
        template = self._template(template_name)
        await self._send_template(template)

    async def _send_template(self, template: Template):
        await self._send_msg(self._chat_id, template)
        if not self._always_retry:
            await self._send_retry_btn(template.name)

    async def _always_retry_dispatcher(self):
        while True:
            message = await self._tw_queue.get()
            if message is None:
                break

            if self._always_retry and self._chat_id:
                action, template_name = message
                template = self._template(template_name)
                await self._send_template(template)

    async def _send_retry_btn(self, template_name: str):
        template = self._template("retry_btn", is_internal=True)
        template.buttons = [
            InlineButton(template_name, self._template_callback(template_name))
        ]
        await self._send_msg(self._chat_id, template)



import asyncio
import os
from pathlib import Path

from aiogram import Bot, Dispatcher, F, Router
from aiogram.client.default import DefaultBotProperties
from aiogram.enums import ParseMode
from aiogram.filters import Command
from aiogram.types import (
    CallbackQuery,
    Message,
)

from live_template.core.router.base_router import BaseRouter
from live_template.core.storage.storage import Template

from .utils import callback_wrapper, make_callback_class, to_message


class AiogramRouter(BaseRouter, Router):
    TEMPLATE_CALLBACK_DATA = make_callback_class(BaseRouter.TEMPLATE_CALLBACK_ID)

    def __init__(self, templates_dir: str | Path, default_template: dict):
        super().__init__(templates_dir, default_template)
        self._bot = None
        self._setup_handlers()

    async def _send_msg(self, chat_id: int, template: Template):
        await self._bot.send_message(self._chat_id, **to_message(template))

    def _setup_handlers(self):
        @self.startup()
        async def on_startup(bot: Bot):
            self._bot = bot
            self._on_startup()

        self.shutdown()(self._on_shutdown)

        @self.message(Command(AiogramRouter.START_COMMAND))
        async def start(message: Message):
            await self._start(message.chat.id)

        for command in self.command_handlers:
            self.message(Command(command))(self.command_handlers[command])

        for callback in self.callback_handlers:
            self.callback_query(F.data == callback)(
                callback_wrapper(self.callback_handlers[callback])
            )

        @self.callback_query(AiogramRouter.TEMPLATE_CALLBACK_DATA.filter())
        @callback_wrapper
        async def get_template(
            callback_query: CallbackQuery,
            callback_data: AiogramRouter.TEMPLATE_CALLBACK_DATA,
        ):
            await self._get_template(callback_data.name)


async def main():
    from dotenv import load_dotenv

    load_dotenv()
    bot_token = os.getenv("BOT_TOKEN")
    bot = Bot(token=bot_token, default=DefaultBotProperties(parse_mode=ParseMode.HTML))
    dp = Dispatcher()
    lt_router = AiogramRouter("../templates", {})
    dp.include_router(lt_router)

    await bot.delete_webhook(drop_pending_updates=True)
    await dp.start_polling(bot)


if __name__ == "__main__":
    asyncio.run(main())



import functools

from aiogram.filters.callback_data import CallbackData
from aiogram.types import (
    CallbackQuery,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
)
from aiogram.utils.keyboard import InlineKeyboardBuilder

from live_template.core.storage.storage import Template


def make_callback_class(prefix: str) -> type:
    class _Callback(CallbackData, prefix=prefix):
        name: str

    return _Callback


def render_buttons(buttons: list, row_sizes: list) -> InlineKeyboardMarkup:
    ikb = InlineKeyboardBuilder()
    if buttons:
        if all(isinstance(b, list) for b in buttons):
            # Flow for full mapped buttons
            for btn_row in buttons:
                ikb.row(*[InlineKeyboardButton(**btn.to_dict()) for btn in btn_row])
        else:
            # Flow for adjusted buttons
            for btn in buttons:
                ikb.button(**btn.to_dict())

            if row_sizes:
                ikb.adjust(*row_sizes)
            else:
                ikb.adjust(2)

    return ikb.as_markup()


def to_message(template: Template) -> dict:
    return {
        "text": template.text,
        "parse_mode": template.parse_mode,
        "reply_markup": render_buttons(template.buttons, template.btn_row_sizes),
    }


def callback_wrapper(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs) -> None:
        callback_query = None
        for arg in args:
            if isinstance(arg, CallbackQuery):
                callback_query = arg
                break
        try:
            await func(*args, **kwargs)
        finally:
            if callback_query:
                await callback_query.answer()

    return wrapper



import ast
from collections.abc import Iterable
from importlib.abc import Traversable
from pathlib import Path

from ..core import log


def get_path_by_name(template_name: str, templates_dir: str | Path) -> Path:
    name = Path(template_name).with_suffix(".py")
    path = Path(templates_dir) / name
    if not path.is_file():
        raise ValueError(f"Wrong path to template '{path}'")
    return path


def get_name_by_path(path: str | Path, templates_dir: str | Path) -> str:
    name = Path(path)
    if not name.suffix == ".py":
        raise ValueError(f"Wrong path to template '{path}'")

    root = Path(templates_dir)
    name = name.relative_to(root).with_suffix("")

    return str(name)


def parse_template(path: str | Path | Traversable) -> dict:
    if not isinstance(path, Traversable):
        path = Path(path)

    try:
        source = path.read_text(encoding="utf-8")
        tree = ast.parse(source)

        data = {}
        for node in tree.body:
            if isinstance(node, ast.Assign):
                target = node.targets[0]
                if isinstance(target, ast.Name) and target.id in {
                    "text",
                    "parse_mode",
                    "buttons",
                    "btn_row_sizes",
                }:
                    data[target.id] = ast.literal_eval(node.value)
        return data
    except (OSError, SyntaxError):
        log.exeption(f"Failed to parse template '{path}'")
        return {}


def get_internal_name_by_path(path: Traversable, templates_dir: Traversable) -> str:
    if not path.name.endswith(".py"):
        raise ValueError(f"Wrong path to template '{path}'")

    if not str(path).startswith(str(templates_dir)):
        raise ValueError(f"Wrong root '{templates_dir}' for path '{path}'")

    return str(path)[len(str(templates_dir)) :].lstrip("/").rstrip(".py")


def get_path_by_internal_name(
    template_name: str, templates_dir: Traversable
) -> Traversable:
    path = templates_dir.joinpath(template_name + ".py")
    if not path.is_file():
        raise ValueError(f"Wrong path to template '{path}'")

    return path


def iter_traversable(path: Traversable, pattern: str = "") -> Iterable:
    if path.is_file() and path.name.endswith(pattern):
        yield path
    elif path.is_dir():
        for child in path.iterdir():
            yield from iter_traversable(child)



from pathlib import Path

from live_template.core.storage.parser import TemplateParser

from .utils import to_message


class AiogramParser(TemplateParser):
    def __init__(self, templates_dir: str | Path):
        super().__init__(templates_dir)

    def get_message(self, template_name: str, **kwargs) -> dict:
        template = self.render_template(template_name, **kwargs)
        if not template:
            raise

        return to_message(template)



text = """
<b>Hello!</b> 👋 I'm a Live Template Package, ready to assist you. Here are some commands you can use:

• <b>/start_lt</b> - to get here
• <b>/list_template_names</b> - to get list of available templates
"""

parse_mode = "HTML"

buttons = [
    {"text": "Templates", "callback_data": "list_template_names"},
    {"text": "Show all", "callback_data": "list_templates"},
    {"text": "Always retry", "callback_data": "toggle_always_retry"},
]

btn_row_sizes = [1]



"""
This module contains TemplateStorage class that implements key-value storage where key is name of
template and value is object of Template class. Storage deal with template names and
translate it to file names to check exists.
"""

from dataclasses import asdict, dataclass
from importlib.abc import Traversable
from pathlib import Path

from ..core import log, package_root
from ..utils.utils import (
    get_internal_name_by_path,
    get_name_by_path,
    get_path_by_internal_name,
    get_path_by_name,
    iter_traversable,
    parse_template,
)


@dataclass
class InlineButton:
    text: str
    callback_data: str

    def to_dict(self) -> dict:
        return asdict(self)


PARSE_MODES = {"HTML", "MARKDOWN", "MARKDOWN_V2"}


@dataclass
class Template:
    name: str
    text: str = ""
    parse_mode: str = "HTML"
    buttons: list[list[InlineButton]] | list[InlineButton] | None = None
    btn_row_sizes: list[int] | None = None

    def to_dict(self) -> dict:
        return asdict(self)

    def __post_init__(self):
        if not (isinstance(self.parse_mode, str) and self.parse_mode in PARSE_MODES):
            log.exception(
                f"Unacceptable value for parse_mode: {self.parse_mode}\n"
                f"Acceptable values: "
                f"{' | '.join(PARSE_MODES)}"
            )
            return

        # Convert buttons from a one- or two-dimensional list of dicts to
        # one- or two-dimensional list of InlineButtons
        def to_inline_button(obj: dict | InlineButton) -> InlineButton:
            return InlineButton(**obj) if isinstance(obj, dict) else obj

        if self.buttons:
            try:
                if all(isinstance(b, list) for b in self.buttons):
                    self.buttons = [
                        [to_inline_button(btn) for btn in row] for row in self.buttons
                    ]
                elif all(isinstance(b, dict) for b in self.buttons):
                    self.buttons = [to_inline_button(btn) for btn in self.buttons]
                else:
                    raise ValueError

            except ValueError:
                log.exception(f"Unacceptable value for buttons: {self.buttons}\n")


class BaseTemplateStorage:
    def __init__(self, templates_dir: str | Path, default_template: dict):
        self._storage: dict[str, Template] = {}
        self.default_template = default_template
        self.templates_dir = Path(templates_dir)

    def add_or_update_template(self, template_name: str):
        path = get_path_by_name(template_name, self.templates_dir)

        try:
            data = parse_template(path)
            for k in ("text", "parse_mode", "buttons", "btn_row_sizes"):
                if k not in data and k in self.default_template:
                    data[k] = self.default_template[k]
            data["name"] = template_name
            template = Template(**data)

            self._storage[template_name] = template

        except Exception:
            log.exception(f"Failed to add or update template '{template_name}'")

    def _load_templates(self):
        for file in self.templates_dir.rglob("*.py"):
            try:
                self.add_or_update_template(get_name_by_path(file, self.templates_dir))
            except ValueError:
                log.exception("Failed to load template")

        log.info("Loaded " + str(self))

    def names(self) -> list:
        return list(self._storage.keys())

    def list(self) -> list:
        return list(self._storage.items())

    def get_template(self, template_name: str) -> Template | None:
        try:
            return self._storage[template_name]
        except KeyError:
            log.exception(f"No template with name '{template_name}'")
            return None

    def __str__(self) -> str:
        return "templates: " + ", ".join(map(lambda s: f"'{s}'", self.names()))

    def __getitem__(self, template_name: str) -> Template:
        return self.get_template(template_name)

    def __contains__(self, template_name: str) -> bool:
        return template_name in self._storage


class TemplateStorage(BaseTemplateStorage):
    def __init__(self, templates_dir: str | Path, default_template: dict):
        super().__init__(templates_dir, default_template)

        if not self.templates_dir.is_dir():
            raise RuntimeError(
                f"Package cannot work without templates directory. Bad path:"
                f" {self.templates_dir.resolve()}"
            )

        self._load_templates()

    def rename_template(self, old_template_name: str, new_template_name: str):
        if old_template_name in self._storage:
            if new_template_name not in self._storage:
                self._storage[new_template_name] = self._storage[old_template_name]
                del self._storage[old_template_name]
            else:
                log.warning(f"New template name '{new_template_name}' already exists!")
        else:
            log.warning(f"Old template name '{old_template_name}' doesn't exists!")

    def delete_template(self, template_name: str):
        if template_name in self._storage:
            del self._storage[template_name]
        else:
            log.warning(f"Template '{template_name}' doesn't exists!")


class InternalTemplateStorage(BaseTemplateStorage):
    def __init__(self, templates_dir: Traversable, default_template: dict):
        super().__init__("", default_template)
        self.templates_dir = templates_dir

        if not self.templates_dir.is_dir():
            raise RuntimeError("Internal Storage bad templates directory!")

        self._load_templates()

    def _load_templates(self):
        for file in iter_traversable(self.templates_dir):
            try:
                self._add(get_internal_name_by_path(file, self.templates_dir))
            except ValueError:
                log.exception("Failed to load template")

        log.info("Loaded " + str(self))

    def _add(self, template_name: str):
        try:
            data = parse_template(
                get_path_by_internal_name(template_name, self.templates_dir)
            )
            for k in ("text", "parse_mode", "buttons", "btn_row_sizes"):
                if k not in data and k in self.default_template:
                    data[k] = self.default_template[k]
            data["name"] = template_name
            template = Template(**data)

            self._storage[template_name] = template
        except Exception:
            log.exception(f"Failed to add or update template '{template_name}'")


def main():
    TemplateStorage("../templates", {})
    InternalTemplateStorage(package_root.joinpath("templates"), {})


if __name__ == "__main__":
    main()



from pathlib import Path

from .storage import Template, TemplateStorage


class TemplateParser:
    def __init__(self, templates_dir: str | Path):
        self._storage = TemplateStorage(templates_dir, {})

    def get_template(self, template_name: str) -> dict:
        template = self._storage[template_name]
        if template:
            return template.to_dict()
        return {}

    def render_template(self, template_name: str, **kwargs) -> Template:
        """
        :param template_name: name of template. user / to separate parts
        :param kwargs: parameters for template text formating
        :return: Template object
        """
        template = self._storage[template_name]
        if template and template.text:
            template.text = template.text.format(**kwargs)

        return template

    def set_storage(self, storage: TemplateStorage):
        """
        WARNING!!! MAY BE UNSTABLE! DO NOT USE IN PRODUCTION ENVIRONMENT!
        :param storage: your lt router storage with realtime updatable templates
        :return:
        """
        self._storage = storage

    def __getitem__(self, item):
        return self.get_template(item)


if __name__ == "__main__":
    tp = TemplateParser("../templates")
    print(tp["dirrrrr/default_template_with_adjusted_buttons"])
    print(tp["ololo"])



import asyncio
import functools
from pathlib import Path
import threading
import time

from watchdog.events import FileSystemEvent, PatternMatchingEventHandler
from watchdog.observers import Observer

from ..core import config, log
from ..storage.storage import TemplateStorage
from ..utils.utils import get_name_by_path


def handle_event(func):
    @functools.wraps(func)
    def wrapper(self, event: FileSystemEvent):
        template_name = get_name_by_path(
            event.src_path, self.watcher.template_storage.templates_dir
        )

        now = time.time()
        last = self.last_changed.get(template_name, 0)
        if now - last < self.debounce:
            return

        self.last_changed[template_name] = now

        return func(self, template_name, event=event)

    return wrapper


class TemplateHandler(PatternMatchingEventHandler):
    def __init__(self, watcher: "TemplateWatcher"):
        super().__init__(patterns=["*.py"], ignore_directories=True)
        self.watcher = watcher
        self.last_changed = {}
        self.debounce = 0.5  # seconds

    @handle_event
    def on_modified(self, template_name: str, **kwargs):
        self.watcher.template_storage.add_or_update_template(template_name)
        log.info(f"Template '{template_name}' was changed")
        if config["always_retry"]:
            asyncio.run_coroutine_threadsafe(
                self.watcher.queue.put(("MODIFIED", template_name)), self.watcher.loop
            )

    @handle_event
    def on_created(self, template_name: str, **kwargs):
        self.watcher.template_storage.add_or_update_template(template_name)
        log.info(f"Template '{template_name}' was added")

    @handle_event
    def on_moved(self, template_name: str, event: FileSystemEvent, **kwargs):
        old_template_name = template_name
        try:
            dest_path = Path(event.dest_path)
            new_template_name = get_name_by_path(
                dest_path, self.watcher.template_storage.templates_dir
            )
            self.watcher.template_storage.rename_template(
                old_template_name, new_template_name
            )
            log.info(
                f"Template '{old_template_name}' was renamed to '{new_template_name}'"
            )
        except ValueError:
            self.watcher.template_storage.delete_template(template_name)
            log.info(f"Template '{template_name}' was removed")

    @handle_event
    def on_deleted(self, template_name: str, **kwargs):
        self.watcher.template_storage.delete_template(template_name)
        log.info(f"Template '{template_name}' was removed")


class TemplateWatcher:
    def __init__(
        self,
        storage: TemplateStorage,
        queue: asyncio.Queue,
        loop: asyncio.AbstractEventLoop,
    ):
        self.template_storage = storage
        self.queue = queue
        self.loop = loop
        self.observer = None

    def start_watching(self):
        self.observer = Observer()
        self.observer.schedule(
            TemplateHandler(self), self.template_storage.templates_dir, recursive=True
        )
        observer_thread = threading.Thread(target=self.observer.start, daemon=True)
        observer_thread.start()
        log.info(
            f"Started '{self.template_storage.templates_dir.resolve()}' observing..."
        )


async def main():
    storage = TemplateStorage("../templates", {})
    queue = asyncio.Queue()
    loop = asyncio.get_running_loop()

    tw = TemplateWatcher(storage, queue, loop)
    tw.start_watching()

    await asyncio.Event().wait()


if __name__ == "__main__":
    asyncio.run(main())



text = """
<b>Hello!</b> 👋 I'm a bot, ready to assist you. Here are some commands you can use:

• <b>/start</b> - start the conversation with the bot 🚀  
• <b>/help</b> - get help with using the bot ❓  
• <b>/info</b> - learn more about the bot ℹ️

   <i>If you have any questions, just let me know!</i> 
"""

parse_mode = "HTML"

buttons = [
    {"text": "Button1_1", "callback_data": "b1_1"},
    {"text": "Button1_2", "callback_data": "b1_2"},
    {"text": "Button1_3", "callback_data": "b1_3"},
    {"text": "Button2_1", "callback_data": "b2_1"},
]

btn_row_sizes = [3, 1]
